@@ -111,16 +111,11 @@ If nil, use the value of `vc-diff-switches'.  If t, use no switches."
 		 (string :tag "Argument String")
 		 (repeat :tag "Argument List" :value ("") string)))
 
-(defun vc-jj--filename-to-fileset (filename &optional root)
-  "Convert FILENAME to a JJ fileset expression.
-If ROOT is non-nil, the resulting fileset is relative to the repository
-root, otherwise it is relative to the current directory."
-  (if root
-      (concat "root-file:\""
-              (string-replace "\"" "\\\"" (file-relative-name filename root))
-              "\"")
-    (concat "cwd-file:\""
-            (string-replace "\"" "\\\"" (file-relative-name filename)) "\"")))
+(defun vc-jj--filename-to-fileset (filename)
+  "Convert FILENAME to a jj fileset expression.
+The fileset expression returned is relative to the jj repository root."
+  (when-let ((path (file-relative-name filename (vc-jj-root filename))))
+    (concat "root:\"" (string-replace "\"" "\\\"" path) "\"")))
 
 (defun vc-jj--process-lines (&rest args)
   "Run jj with ARGS, returning its output to stdout as a list of strings.
@@ -195,33 +190,61 @@ stderr and1 `vc-do-command' cannot separate output to stdout and stderr."
            (/= (point-min) (point-max))))))
 
 (defun vc-jj-state (file)
-  "JJ implementation of `vc-state' for FILE."
-  ;; We need to run two commands for the complete state:
+  "JJ implementation of `vc-state' for FILE.
+There are several file states recognized by vc (see the docstring of
+`vc-state' for the full list).  Several of these are relevant to
+jujutsu.  They are:
+- Added (new file)
+- Removed (deleted file)
+- Edited (modified file)
+- Up-to-date (unmodified file)
+- Conflict (merge conflict)
+- Ignored (ignored by repository)
+
+Other VCS backends would also recognize the \"unregistered\" state, but
+there is no such state in jj since jj automatically registers new files."
+  ;; We can deduce all of the vc states listed above with two
+  ;; commands:
   ;;
-  ;; - "jj file list -T 'conflict' FILE" gets us conflicted (output
-  ;;   "true"), ignored (no output) or tracked (output "false", but
-  ;;   could be added or modified)
+  ;; - "jj diff --summary FILE" gets us the "added" state (when output
+  ;;   starts with "A "), the "removed" state (when output starts with
+  ;;   "D ") and "edited" state (when output starts with "M ").  No
+  ;;   output is also possible (this could mean the "conflict",
+  ;;   "ignored" or "unchanged" state), but we deduce these with the
+  ;;   command below instead.
+  ;; 
+  ;; - "jj file list -T 'conflict' FILE" gets us the "conflict" state
+  ;;   (when output is "true"), the "ignored" state (when there is no
+  ;;   output) and states for tracked files (when output is "false",
+  ;;   either the "added", "removed", "edited" or "up-to-date" state).
+  ;;   We deduce the "added", "removed" and "edited" states with the
+  ;;   command above.
   ;;
-  ;; - "jj diff --summary FILE" gets us modified (output starts with
-  ;;   "M ") or added (output starts with "A "), but no output could
-  ;;   be conflicted, ignored or unchanged
+  ;;   The only remaining possibility is the "up-to-date" state, which
+  ;;   we can deduce because when all other states have not been
+  ;;   matched against.
   (let* ((default-directory (vc-jj-root file))
          (file (vc-jj--filename-to-fileset file))
+         (changed
+          (vc-jj--command-parseable "diff" "--summary" "--" file))
          (conflicted-ignored
-          (vc-jj--command-parseable "file" "list" "-T" "conflict" "--" file))
-         (modified-added
-          (vc-jj--command-parseable "diff" "--summary" "--" file)))
+          (vc-jj--command-parseable "file" "list" "-T" "conflict" "--" file)))
     (cond
-     ((string-empty-p conflicted-ignored) 'ignored)
+     ((string-prefix-p "A " changed) 'added)
+     ((string-prefix-p "D " changed) 'removed)
+     ((string-prefix-p "M " changed) 'edited)
      ((string= conflicted-ignored "true") 'conflict)
-     ((string-prefix-p "M " modified-added) 'edited)
-     ((string-prefix-p "A " modified-added) 'added)
+     ((string-empty-p conflicted-ignored) 'ignored)
+     ;; If the file hasn't matched anything yet, this leaves only one
+     ;; possible state: up-to-date
      ((string= conflicted-ignored "false") 'up-to-date)
-     (t nil))))
+     (t (error "VC state of %s is not recognized" file)))))
 
 (defun vc-jj-dir-status-files (dir _files update-function)
   "Calculate a list of (FILE STATE EXTRA) entries for DIR.
-Return the result result of applying UPDATE-FUNCTION to that list."
+Return the result of applying UPDATE-FUNCTION to that list.
+
+For a description of the states relevant to jj, see `vc-jj-state'."
   ;; This function is specified below the STATE-QUERYING FUNCTIONS
   ;; header in the comments at the beginning of vc.el.  The
   ;; specification says the 'dir-status-files' backend function
@@ -239,41 +262,47 @@ Return the result result of applying UPDATE-FUNCTION to that list."
   ;;
   ;; TODO: we should use hash tables, since we're doing a lot of set
   ;; operations, which are slow on lists.
-  (let* ((dir (expand-file-name dir))
-         (default-directory dir)
-         (project-root (vc-jj-root dir))
-         (registered-files (vc-jj--process-lines "file" "list" "--" dir))
-         (ignored-files (seq-difference (cl-delete-if #'file-directory-p
-                                                      (directory-files dir nil nil t))
-                                        registered-files))
-         (changed (vc-jj--process-lines "diff" "--summary" "--" dir))
-         (added-files (mapcan (lambda (entry)
-                                (and (string-prefix-p "A " entry)
-                                     (list (substring entry 2))))
-                              changed))
-         (modified-files (mapcan (lambda (entry)
-                                   (and (string-prefix-p "M " entry)
-                                        (list (substring entry 2))))
-                                 changed))
-         ;; The command below only prints conflicted files in DIR, but
-         ;; relative to project-root, hence the dance with
-         ;; expand-file-name / file-relative-name
-         (conflicted-files (mapcar (lambda (entry)
-                                     (file-relative-name (expand-file-name entry project-root) dir))
-                                   (vc-jj--process-lines "file" "list"
-                                                         "-T" "if(conflict, path ++ \"\\n\")" "--" dir)))
-         (unchanged-files (cl-remove-if (lambda (entry) (or (member entry conflicted-files)
-                                                            (member entry modified-files)
-                                                            (member entry added-files)
-                                                            (member entry ignored-files)))
-                                        registered-files))
-         (result
-          (nconc (mapcar (lambda (entry) (list entry 'conflict)) conflicted-files)
-                 (mapcar (lambda (entry) (list entry 'added)) added-files)
-                 (mapcar (lambda (entry) (list entry 'edited)) modified-files)
-                 (mapcar (lambda (entry) (list entry 'ignored)) ignored-files)
-                 (mapcar (lambda (entry) (list entry 'up-to-date)) unchanged-files))))
-    (funcall update-function result nil)))
+  (with-demoted-errors "JJ error during `vc-dir-status-files': %S"
+    (let* ((dir (expand-file-name dir))
+           (default-directory dir)
+           (project-root (vc-jj-root dir))
+           (registered-files (vc-jj--process-lines "file" "list" "--" dir))
+           (ignored-files (seq-difference (cl-delete-if #'file-directory-p
+                                                        (directory-files dir nil nil t))
+                                          registered-files))
+           (changed (vc-jj--process-lines "diff" "--summary" "--" dir))
+           (added-files (mapcan (lambda (entry)
+                                  (and (string-prefix-p "A " entry)
+                                       (list (substring entry 2))))
+                                changed))
+           (removed-files (mapcan (lambda (entry)
+                                    (and (string-prefix-p "D " entry)
+                                         (list (substring entry 2))))
+                                  changed))
+           (edited-files (mapcan (lambda (entry)
+                                     (and (string-prefix-p "M " entry)
+                                          (list (substring entry 2))))
+                                   changed))
+           ;; The command below only prints conflicted files in DIR, but
+           ;; relative to project-root, hence the dance with
+           ;; expand-file-name / file-relative-name
+           (conflicted-files (mapcar (lambda (entry)
+                                       (file-relative-name (expand-file-name entry project-root) dir))
+                                     (vc-jj--process-lines "file" "list"
+                                                           "-T" "if(conflict, path ++ \"\\n\")" "--" dir)))
+           (up-to-date-files (cl-remove-if (lambda (entry) (or (member entry conflicted-files)
+                                                              (member entry edited-files)
+                                                              (member entry added-files)
+                                                              (member entry ignored-files)))
+                                          registered-files))
+           (result
+            (nconc (mapcar (lambda (entry) (list entry 'conflict)) conflicted-files)
+                   (mapcar (lambda (entry) (list entry 'added)) added-files)
+                   (mapcar (lambda (entry) (list entry 'removed)) removed-files)
+                   (mapcar (lambda (entry) (list entry 'edited)) edited-files)
+                   (mapcar (lambda (entry) (list entry 'ignored)) ignored-files)
+                   (mapcar (lambda (entry) (list entry 'up-to-date)) up-to-date-files))))
+      (funcall update-function result nil))))
 
 (defun vc-jj-dir-extra-headers (dir)
   "Return extra headers for `vc-dir' when executed inside DIR.
@@ -414,6 +443,10 @@ parents.map(|c| concat(
 
 (defun vc-jj-checkin (files comment &optional _rev)
   "Run \"jj commit\" with supplied FILES and COMMENT."
+  ;; FIXME 2025-06-24: Find a more robust solution than
+  ;; `replace-regexp-in-string' on the "Summary" header if possible.
+  ;; Also ideally future-proof by considering the presence of other
+  ;; headers, e.g., `log-edit-headers-alist'.
   (setq comment (replace-regexp-in-string "\\`Summary: " "" comment))
   (let ((args (append (vc-switches 'jj 'checkin) (list "commit" "-m" comment))))
     (apply #'vc-jj--command-dispatched nil 0 files args)))
@@ -422,7 +455,7 @@ parents.map(|c| concat(
   "Read revision REV of FILE into BUFFER and return the buffer."
   (let ((revision (vc-jj--command-parseable "file" "show" "-r" rev "--" (vc-jj--filename-to-fileset file))))
     (with-current-buffer buffer
-      (erase-buffer buffer)
+      (erase-buffer)
       (insert revision)))
   buffer)
 
@@ -455,8 +488,7 @@ If REV is not specified, revert the file as with `vc-jj-revert'."
                  files))))
     (with-current-buffer buffer
       (apply #'vc-jj--command-dispatched buffer
-        'async nil
-        args))))
+        'async nil "log" args))))
 
 (defun vc-jj-show-log-entry (revision)
   "Move to the log entry for REVISION."
@@ -600,12 +632,12 @@ If REV is not specified, revert the file as with `vc-jj-revert'."
   (and vc-parent-buffer
     (with-current-buffer vc-parent-buffer
       (revert-buffer)))
-  (with-current-buffer (revert-buffer)))
+  (revert-buffer))
 
 (defun vc-jj-edit-change ()
   (interactive)
   (let ((rev (log-view-current-tag)))
-    (vc-jj--command-dispatched nil 0 nil "edit" rev "--quiet")
+    (vc-jj-retrieve-tag nil rev nil)
     (vc-jj--reload-log-buffers)))
 
 (defun vc-jj-abandon-change ()
@@ -620,7 +652,6 @@ If REV is not specified, revert the file as with `vc-jj-revert'."
   (interactive)
   (let ((rev (log-view-current-tag)))
     (vc-jj--command-dispatched nil 0 nil "new" rev "--quiet")
-    (with-current-buffer (revert-buffer))
     (vc-jj--reload-log-buffers)))
 
 (defun vc-jj-root (file)
@@ -721,15 +752,39 @@ four groups: change id, author, datetime, line number.")
     (when (looking-at vc-jj--annotation-line-prefix-re)
       (match-string-no-properties 1))))
 
+(defun vc-jj--revision-annotation-function (elem)
+  "Calculate propertized change description from ELEM.
+ELEM can be of the form (change-id . description), as produced in
+`vc-jj-revision-completion-table', in which case we return the second
+element, or can be a change id, in which case we query for the first
+line of its description."
+  (let ((description
+         (if (listp elem)
+             (second elem)
+           (vc-jj--command-parseable "log" "-r" elem "--no-graph" "-T" "self.description().first_line()"))))
+    (format " %s" (propertize description 'face 'completions-annotations))))
+
 (defun vc-jj-revision-completion-table (files)
-  (let ((revisions
-         (apply #'vc-jj--process-lines "log" "--no-graph"
-                "-T" "self.change_id() ++ \"\\n\"" "--" files)))
+  "Return a completion table for existing revisions of FILES."
+  (let* ((revisions
+          (mapcar
+           ;; Boldly assuming that jj's change ids won't suddenly change length
+           (lambda (line) (list (substring line 0 31) (substring line 32)))
+           (apply #'vc-jj--process-lines "log" "--no-graph"
+                  "-T" "self.change_id() ++ self.description().first_line() ++ \"\\n\"" "--" files))))
     (lambda (string pred action)
       (if (eq action 'metadata)
-          `(metadata . ((display-sort-function . ,#'identity)))
+          `(metadata . ((display-sort-function . ,#'identity)
+                        (annotation-function . ,#'vc-jj--revision-annotation-function)))
         (complete-with-action action revisions string pred)))))
 
+(defun vc-jj-retrieve-tag (_dir rev _update)
+  "Call jj edit on REV inside DIR.
+REV is the change ID of a jj revision.
+
+_DIR and _UPDATE are as described in the vc.el specification."
+  (vc-jj--command-dispatched nil 0 nil "edit" rev "--quiet"))
+
 (defvar vc-jj-pull-history nil
   "History variable for `vc-jj-pull'.")
 
